'''
This is where we store standard examples of pairwise and supermatrices.

'''
import pandas as pd
from pyanp.general import islist

# The dataframe storing the pairwise example matrices
PAIRWISE_EXS = pd.DataFrame(columns=["description", "keywords", "matrix", "size", "author"])
# The dataframe storing the supermatrix examples matrices
SUPERMATRIX_EXS = pd.DataFrame(columns=["description", "keywords", "matrix", "size", "author"])


def _clean_keywords(keywords):
    '''
    Cleans keywords for example matrices, to make sure they are easily searched
    :param keywords: The keywords sent to '_add_pairwise' or '_add_supermatrix'
    :return: A cleansed version, stripped and to lowercase, among other things.
    '''
    if keywords is None:
        return []
    if isinstance(keywords, str):
        #Only one keyword
        return [keywords.strip().lower()]
    elif islist(keywords):
        return [_clean_keywords(keyword) for keyword in keywords]
    else:
        return keywords


def _add_pairwise(name, description, keywords, mat, author=''):
    '''
    Adds a new pariwise example
    :param name: The unique id of this pairwise matrix example.
    :param description: A stringy description of the example
    :param keywords: A list of stringy keywords
    :param mat: The numpy matrix
    :param author: Optional author email
    :return:
    '''
    global PAIRWISE_EXS
    if name in PAIRWISE_EXS.index:
        raise ValueError("Already had a pairwise example with that name")
    size = len(mat)
    keywords = _clean_keywords(keywords)
    PAIRWISE_EXS.loc[name]=[description, keywords, mat, size, author]

def _add_supermatrix(name, description, keywords, mat, author=''):
    '''
    Adds a new supermatrix example
    :param name: Unique string id of the example
    :param description: A string description
    :param keywords: A list of string keywords
    :param mat: The numpy matrix
    :param author: Optional string email of the contributor/author
    :return:
    '''
    global SUPERMATRIX_EXS
    if name in SUPERMATRIX_EXS.index:
        raise ValueError("Already had a supermatrix example with that name")
    size = len(mat)
    keywords = _clean_keywords(keywords)
    SUPERMATRIX_EXS.loc[name]=[description, keywords,  mat, size, author]


def matrix_matching(df=SUPERMATRIX_EXS, description=None, keywords=None, size=None, author=None):
    '''
    Finds matrices that match search criteria
    :param df: The dataframe to search through, either SUPERMATRIX_EXS or PAIRWISE_EXS
    :param description: A substring to search through description
    :param keywords: A list of keywords to find
    :param size: The size
    :param author: The contributing author, a substring search
    :return: List of indices of the matches in the given dataframe
    '''
    indices = list(df.index)
    if description is not None:
        description = description.lower()
        indices = [index for index in indices if description in
                   df.loc[index, 'description'].lower()]
    if keywords is not None:
        if islist(keywords):
            for keyword in keywords:
                keyword = str(keyword).lower().strip()
                indices = [i for i in indices if keyword in df.loc[i, 'keywords']]
        elif isinstance(keywords, str):
            keywords = str(keywords).lower().strip()
            indices = [i for i in indices if keywords in df.loc[i, 'keywords']]
    if size is not None:
        indices = [i for i in indices if size == df.loc[i, 'size']]
    if author is not None:
        author = author.strip().lower()
        indices = [i for i in indices if author in df.loc[i, 'keywords']]
    return indices


def supermatrix_ex(name=None, description=None, keywords=None, size=None, author=None):
    '''
    Find the supermatrix example that matches the conditions
    :param name: If not None, we find the single matrix with this name/id
    :param description: Substring search in description
    :param keywords: exact match keywords
    :param size: exact match size
    :param author: substring search author
    :return: A single numpy.array item if only one matches the constraint, or a panda.Series
    indexed by name or the resulting numpy.array s.
    '''
    global SUPERMATRIX_EXS
    if name is not None:
        return SUPERMATRIX_EXS.loc[name,'matrix']
    else:
        indices = matrix_matching(SUPERMATRIX_EXS, description, keywords, size, author)
        return SUPERMATRIX_EXS.loc[indices, 'matrix']

def pairwisematrix_ex(name=None, description=None, keywords=None, size=None, author=None):
    '''
    Find the pairwise matrix example that matches the conditions
    :param name: If not None, we find the single matrix with this name/id
    :param description: Substring search in description
    :param keywords: exact match keywords
    :param size: exact match size
    :param author: substring search author
    :return: A single numpy.array item if only one matches the constraint, or a panda.Series
    indexed by name or the resulting numpy.array s.
    '''
    global PAIRWISE_EXS
    if name is not None:
        return PAIRWISE_EXS.loc[name,'matrix']
    else:
        indices = matrix_matching(PAIRWISE_EXS, description, keywords, size, author)
        return PAIRWISE_EXS.loc[indices, 'matrix']




from pyanp.exmatcontrib import *


'''
Generally useful math and other functions.
'''

def linear_interpolate(xs, ys, x):
    '''

    :param xs: Increasing x values (no dupes)
    :param ys: The y values
    :param x: The x value to linearly interpolate at
    :return: if x < xs[0], returns ys[0], if x > xs[-1], returns ys[-1]
    else linearly interpolates.
    '''
    if x <= xs[0]:
        return ys[0]
    elif x >= xs[-1]:
        return ys[-1]
    #Okay we are in between, find first xs we are in between
    for i in range(1, len(xs)):
        if x <= xs[i]:
            slope = (ys[i]-ys[i-1])/(xs[i]-xs[i-1])
            return ys[i-1]+(x-xs[i-1])*slope
    #Should never make it here
    raise ValueError("linear interpolation failure")


def islist(val):
    '''
    Simple function to check if a value is list like object
    :param val:
    :return:
    '''
    if val is None:
        return False
    elif isinstance(val, str):
        return False
    else:
        return hasattr(val, "__len__")




'''
All pairwise matrix to priority vector calculations

@author: Dr. Bill Adams
'''
import numpy as np

######################################################
#### Priority Vector Calculations                 ####
######################################################

def geom_avg(vals):
    """
    Compute the geometric average of a list of values.
    
    The values need not be a list, but simply anything with a len() and []
    """
    rval=1.0
    count = 0
    for val in vals:
        val = vals[count]
        if val != 0:
            rval *= val
            count+=1
    if count != 0:
        rval = pow(rval, 1.0/count)
    return(rval)

def geom_avg_mat(mat, coeffs = None):
    '''
    Computes the geometric average of the columns of a matrix.  Returns
    an np.array of dimension [nRowsOfMat], i.e. a vector.  
    
    :param mat: Must be an numpy.array of shape [nRows, nCols]
    :param coeffs:  If not None, it is a list like object with nColsOfMat elements.
    We multiply column 0 of mat by coeffs[0], column 1 of mat by coeffs[1], etc
    and then do the geometric average of the columns.  Essentially this weights the
    columns.
    '''
    size = mat.shape[0]
    rval = np.ones([size])
    # Normalize the coeffs if they exist
    if np.any(coeffs):
        coeffs = size*np.array(coeffs)/sum(coeffs)
    for row in range(size):
        if np.any(coeffs):
            theRow = mat[row,:] ** np.array(coeffs)
        else:
            theRow = mat[row,:]
        rval[row] = geom_avg(theRow)
    return(rval)

def pri_expeigen(mat, error = 1e-10):
    """
    Calculates priorities using exponential (aka multiplicative eigenvector)

    :param mat: An numpy.array of shape [size, size] of pairwise comparisions.
    :param error=1e-10: The convergence error term
    :return numpy.array: The resulting exponential eigenvector as a numpy.array of shape [size]
    """
    size = mat.shape[0]
    vec = np.ones([size])
    diff = 1
    count=0
    while diff >= error and count < 100:
        nextv = geom_avg_mat(mat, vec)
        nextv = nextv/max(nextv)
        diff = max(abs(nextv - vec))
        vec = nextv
        count+=1
    return(vec/sum(vec))

def pri_llsm(mat):
    '''
    Calculates the priorities using the geometric mean method, aka Log Least Squares Method (LLSM).
    :param mat: An numpy.array of dimension [size,size]
    :return numpy.array: The resulting llsm priority vector as a numpy.array of shape [size]
    '''
    rval = geom_avg_mat(mat)
    rval = rval / sum(rval)
    return(rval)

# LLSM is the same geometric average, so let's make a synonym
pri_geomavg = pri_llsm

def harker_fix(mat):
    """
    Performs Harker's fix on the numpy matrix mat.  It returns a copy with the fix.
    The function does not change the matrix mat.
    :param mat: A numpy array
    :return: A copy of mat with Harker's fix applied to it
    """
    nrows = mat.shape[0]
    ncols = mat.shape[1]
    rval = mat.copy()
    for row in range(nrows):
        val = 1
        for col in range(ncols):
            if col != row and mat[row,col]==0:
                val+=1
        rval[row,row]=val
    return(rval)


def pri_eigen(mat, error = 1e-10, use_harker = False, return_eigenval=False):
    '''
    Calculates the largest eigen vector of a matrix
    
    :param mat: A square numpy array.
    :param use_harker=False: Should we apply Harker's fix before computing?
    :param return_eigenval=False: If True it returns only the eigenvalue, otherwise only returns the eigenvector.
    :return numpy.array: The largest eigenvector that is the normalized (sum to 1) largest eigenvector as a numpy.array
        of shape [size] if return_eigenval=False, otherwise returns the eigenvalue as a number.
    '''
    if use_harker:
        mat = harker_fix(mat)
    size = mat.shape[0]
    vec = np.ones([size])
    diff = 1
    while diff > error:
        nextv = np.matmul(mat, vec)
        nextv = nextv/sum(nextv)
        diff = max(abs(nextv - vec))
        vec = nextv
    if return_eigenval:
        nextv = np.matmul(mat, vec)
        return sum(nextv)
    else:
        return(vec)


def inconsistency_divisor(mat_or_size):
    '''
    Calculates the inconsistency divisor for a matrix, or the size of a matrix.

    The inconsistency divisor is what you divid (eigenvalue - size) by to get the inconsistency.
    :param mat_or_size: Either a pairwise matrix, or simply the size of the pairwise matrix (which is
    what determines the inconsistency divisor).
    :return: The inconsistency divisor
    '''
    size = size_array_like(mat_or_size)
    t=size - 1
    if size<=0:
        return 1
    elif size==1:
        return 1
    elif size==2:
        return 1
    elif size==3:
        return .52 * t
    elif size==4:
        return .89 * t
    elif size==5:
        return 1.12 * t
    elif size==6:
        return 1.25 * t
    elif size==7:
        return 1.35 * t
    elif size==8:
        return 1.40 * t
    elif size==9:
        return 1.45 * t
    elif size==10:
        return 1.49 * t
    elif size==11:
        return 1.51 * t
    elif size==12:
        return 1.54 * t
    elif size==13:
        return 1.56 * t
    elif size==14:
        return 1.57 * t
    elif size==15:
        return 1.58 * t
    else:
        return 1.98 * (1 - (size - 1) / (size * (size - 1) / 2))

def incon_std(mat, error = 1e-10, use_harker = False):
    '''
    Calculates the inconsistency of a pairwise matrix using the standard AHP/ANP theoretic formula.

    :param mat: A numpy.array of shape [size,size] of pairwise comparisons.
    :param error: The error to use for the pri_eigen calculation
    :param use_harker: Should we apply Harker's fix before the calculation?
    :return: The inconsistency.
    '''
    size = mat.shape[0]
    largest_eigen_val = pri_eigen(mat, error, use_harker, return_eigenval=True)
    return (largest_eigen_val-size)/inconsistency_divisor(mat)


#########################################################
### Priority Error Calculations                  ########
#########################################################
def prerr_euclidratio(pwmat, privec):
    '''
    Calculates the euclidean distance error between the pairwise matrix and the ratio matrix of a priority vector.

    This calculates using the following formula
    sqrt( sum_{i,j} (  (pwmat[i, j] - privec[i]/privec[j])**2 ))
    :param pwmat: A numpy.array of shape [size, size] of pairwise comparisons.
    :param privec: A numpy.array of share [size] of the priority vector
    :return:
    '''
    rval = 0
    diffsum = 0
    count = 0
    size = pwmat.shape[0]
    for i in range(0, size):
        for j in range(0, size):
            if (i != j) and (pwmat[i, j] != 0):
                diffsum += (pwmat[i, j] - privec[i] / privec[j]) ** 2
                count += 1
    if count == 0:
        return 0
    else:
        return diffsum ** (1.0 / 2)


def prerr_ratio_avg(pwmat, privec):
    '''
    Calculates priority error using the arithmetic average of ratio distance of pwmat from the ratio matrix of privec

    It averages:
    ratio_greater_1(pwmat[i, j], (privec[i]/privec[j])) - 1
    where ratio_greater_1(a,b) is 1 if a or b is 0 and is max(a/b, b/a) otherwise.

    :param pwmat: A numpy.array of shape [size, size] of pairwise comparisons
    :param privec: A numpy.array of shape [size] of priortiy vector
    :return: The ratio average priority vector
    '''
    diffsum = 0
    count = 0
    size = pwmat.shape[0]
    rmat = ratio_mat(privec)
    for i in range(0, size):
        for j in range(0, size):
            if (pwmat[i, j] >= 1) and (i != j):
                ratio = ratio_greater_1(pwmat[i, j], rmat[i, j])
                score = ratio - 1
                diffsum += score
                count += 1
                # print("ratio={} diffprod={}".format(ratio, diffprod))
    if count == 0:
        return 0
    else:
        return diffsum * (1.0 / count)

def ratio_greater_1(a, b):
    '''
    The ratio of a to b (or b to a) that is larger than or equal to 1.

    :param a: A numerical value for the ratio calculation
    :param b: Another numerical value
    :return: 1 if a or b is 0, otherwise max(a/b, b/a)
    '''
    if (a == 0) or (b == 0):
        return 1
    else:
        return max(a/b, b/a)

def prerr_ratio_prod(pwmat, privec):
    '''
    Calculates priority error using the geometric average of ratios of pwmat and the ratio matrix of privec

    :param pwmat: A numpy.array of shape [size, size] of pairwise comparisons
    :param privec: A numpy.array of shape [size] of priortiy vector
    :return: The calculated error
    '''
    diffprod = 1
    count = 0
    size = pwmat.shape[0]
    for i in range(0, size):
        for j in range(0, size):
            if (pwmat[i, j] >= 1) and (i != j):
                ratio = ratio_greater_1(pwmat[i, j], privec[i]/privec[j])
                diffprod *= ratio
                count += 1
                #print("ratio={} diffprod={}".format(ratio, diffprod))
    #print(diffprod)
    if count == 0:
        return 0
    else:
        return diffprod ** (1.0 / count)

######################################################
### Creating a pairwise matrix from simpler data #####
######################################################

def ratio_mat(pv):
    '''
    Returns the ratio matrix of a vector

    :param pv: An array-like object with len(pv)=size
    :return: A numpy.array of shape [size, size] of the ratios
    '''
    size = len(pv)
    rval = np.identity(n=size)
    for row in range(size):
        for col in range(size):
            if pv[col] != 0:
                rval[row, col] = pv[row] / pv[col]
    return rval

def utmrowlist_to_npmatrix(list_of_votes):
    '''
    Convert a list of values to a pairwise matrix, assuming the list is the upper triangular part only

    :param list_of_votes: An array like, the first elements are the top row of the UTM part of the matrix.
        Then it goes to the second row, etc.
    :return: A numpy.array of the full pairwise comparison matrix.
    '''
    N = len(list_of_votes)
    # Find dims
    n_float = (1 + np.sqrt(1+8*N))/2
    # This should be an integer for it to work
    n = int(n_float)
    if (np.abs(n-n_float) > 1e-10):
        raise ValueError("The list is the wrong size for a utm")
    rval = np.identity(n)
    pos = 0
    for row in range(n):
        for col in range(row+1, n):
            val = list_of_votes[pos]
            rval[row, col] = val
            if val != 0:
                rval[col, row] = 1/val
            pos += 1
    return rval


######################################################
### Helper functions  ################################
######################################################
def size_array_like(mat_or_size):
    '''
    Returns the size of an array like or integer
    :param mat_or_size: Either an integer (specifying the size) or an array-like or numpy.array of shape [size, size].
        If array-like list of lists, we only use len(mat_or_size), we do not check that the array-like is actually
        square.
    :return: The parameter if it was an integer, len(mat_or_size) if param is a list, or mat_or_size.shape[0] if
        mat_or_size is a numpy.array
    '''
    if isinstance(mat_or_size, (int)):
        return mat_or_size
    elif isinstance(mat_or_size, (np.ndarray)):
        return mat_or_size.shape[0]
    elif isinstance(mat_or_size, (tuple, list)):
        return len(mat_or_size)
    else:
        raise ValueError("Unable to get size")




'''
Created October 24, 2017

@author: Bill Adams
'''

from setuptools import setup

setup(name='pyanp',
      version='0.2',
      description='Math library for ANP/AHP calculations',
      url='https://github.com/wjladams/pyanp',
      author='Bill Adams',
      author_email='wjadams@decisionlens.com',
      license='MIT',
      package_dir={'pyanp':'pyanp'},
      python_requires=">=3",
      install_requires=['numpy'],
      packages=['pyanp'],
      #py_modules=['fxmodel'],
      #package_data={'bamath': ['data/*.csv', 'data/*.pickle', 'data/*.json', '*.css']},
      zip_safe=False)



'''
ANP row sensitivity calculations
'''

import pandas as pd
from copy import deepcopy
from pyanp.general import linear_interpolate
from enum import Enum
from pyanp.limitmatrix import calculus, priority_from_limit
from scipy.stats import rankdata
import numpy as np
import matplotlib.pyplot as plt


def _p_to_scalar(mat, p, p0=0.5):
    '''
    Simple internal function that turns a p parameter for anp row sensitivity
    into the scalar value to multiply the row by.
    :param mat: The matrix that this p will be applied to.  As of this writing, it is not used in the calculation.
    :param p: The p-value to convert.  Should be a number between 0 and 1.
    :param p0: The fixed value, should be a number between 0 and 1 EXCLUSIVE
    :return: A list of two items.  The first is a boolean telling whether we scaled Up (True) or Down (False) and the
    second item is the scalar to multiply the row by.
    '''
    if p < p0:
        return (False, p/p0)
    else:
        return (True, (1-p)/(1-p0))



def calcp0(mat, row, cluster_nodes, orig, p0mode):
    '''
    Calculates the p0, or resting, value for the row sensitivity
    :param mat: The matrix to do row sensitivity on
    :param row: The row to do row sensitivity on
    :param cluster_nodes: The indices of the other nodes in the cluster that `row` is in.  Used for
    `inluence_marginal()` if p0mode is an integer meaning smart mode for alt=p0mode
    :param orig: The original weight, used if p0mode is not an integer (meaning smart) or a float (meaning a direct
    p0 value)
    :param p0mode:  This controls the calculation and has 3 cases:
    * if it is a float, you are directly setting the p0 value to whatever p0mode is
    * if it is an integer, this is the smart p0 mode, and it treats p0mode as the index of the alternative/node to make
    continuous.
    * otherwise we assume you want original weights to be the p0 value, and return the parameter `orig`
    :return: The p0 or resting value, see the `p0mode` parameter for more information
    '''
    if isinstance(p0mode, float):
        # p0 has been directly sent
        p0 = p0mode
    elif isinstance(p0mode, int):
        # This is smart mode
        cont_alt = p0mode
        left_deriv = influence_marginal(mat, row, influence_nodes=None, cluster_nodes=cluster_nodes,
                                        left_or_right=-1)
        right_deriv = influence_marginal(mat, row, influence_nodes=None, cluster_nodes=cluster_nodes,
                                        left_or_right=+1)
        lval = left_deriv[cont_alt]
        rval = right_deriv[cont_alt]
        p0 = lval / (lval + rval)
    else:
        # Use the original weight
        p0 = orig
    return p0

def row_adjust(mat, row, p, cluster_nodes=None, inplace=False, p0mode=None):
    '''
    Performs an actual row adjust on the matrix, either inplace or returns an adjusted copy, leaving the
    original unchanged.
    :param mat: The scaled supermatrix to perform anp row sensitivity on.
    :param row: The row index to perform the anp row sensitivity on
    :param p: The p value to adjust to
    :param cluster_nodes: The other nodes in the parameter row's cluster (including row itself),
    so we can scale by cluster.  If None we do not scale by cluster.
    :param inplace: Should we change the matrix mat, or should we create a new one, adjust it, and return it?
    :param p0mode:  See calcp0() function
    :return: The adjusted matrix if inplace=False, and otherwise returns nothing and changes the matrix mat.
    '''
    n = len(mat)
    if not inplace:
        mat = deepcopy(mat)
    if cluster_nodes is None:
        cluster_nodes = range(n)
    if row not in cluster_nodes:
        raise ValueError("Row was not in cluster_nodes, cannot adjust like that")
    for c in range(n):
        # First normalize column across cluster_nodes
        total = sum(abs(mat[cluster_nodes, c]))
        if total != 0:
            mat[cluster_nodes, c] /= total
        #Now we can continue
        orig = mat[row,c]
        if (orig != 0) and (orig != 1):
            # Get our p0
            p0 = calcp0(mat, row, cluster_nodes, orig, p0mode)
            scale_up, scalar = _p_to_scalar(mat, p, p0)
            if not scale_up:
                mat[row,c]*=scalar
                for r in cluster_nodes:
                    if r != row:
                        mat[r,c] *= (1-mat[row,c])/(1-orig)
            else:
                mat[row,c] = 1 - scalar*(1-mat[row,c])
                for r in cluster_nodes:
                    if r != row:
                        mat[r, c] *= scalar
        #Now we normalize back to original sum
        if total != 0:
            mat[cluster_nodes,c] *= total

    if not inplace:
        return mat



def p0mode_name(p0mode_value):
    '''
    Tells what kind of p0 the p0mode value is
    :param p0mode_value:
    :return: A string/human readable bit of information about the p0mode.
    '''
    if isinstance(p0mode_value, int):
        # This is smart p0
        return 'smart'
    elif isinstance(p0mode_value, float):
        # Directvalue
        return 'direct'
    else:
        return 'original weight'

def p0mode_is_smart(p0mode_value):
    '''
    Is the p0mode value a "smart value".  See calcp0 for more info on p0mode values.
    :param p0mode_value:  If it is an int, this p0mode value represents doing smart p0 making the node of that index
    (i.e.p0mode's value) smooth
    :return: True | False
    '''
    return isinstance(p0mode_value, int)

def p0mode_is_direct(p0mode_value):
    '''
    Is the p0mode value a "directly set value".  See calcp0 for more info on p0mode values.
    :param p0mode_value:  If it is an float, this p0mode value is the direct value to use for p0.
    :return: True | False
    '''
    return isinstance(p0mode_value, float)

def influence_marginal(mat, row, influence_nodes=None, cluster_nodes=None, left_or_right=None, delta=1e-6,
                       p0mode=0.5, limit_matrix_calc=calculus):
    '''
    Calculates the marginal influence
    :param mat: The scaled supermatrix to calculate marginal influence on
    :param row: The index of the row to perform the marginal influence on
    :param influence_nodes: The nodes to calculate the marginal influence of the row upon, if None then it assumes
    all nodes except row.
    :param cluster_nodes: The other nodes in the parameter row's cluster (including row itself),
    so we can scale by cluster.  If None we do not scale by cluster.
    :param left_or_right: An integer telling whether we should do left-hand side derivative, right-hand side
    derivative or average them.  If left_or_right < 0, then we do LHS deriv.  If left_or_right > 0, we do RHS deriv.
    Finally, if left_or_right == 0, we average LHS and RHS.
    :param delta: The delta_x to use for the derivative calculation.
    :param p0mode: Determines how p0 is calculated and has 3 cases:
    * if it is a float, you are directly setting the p0 value to whatever p0mode is
    * if it is an integer, this is the smart p0 mode, and it treats p0mode as the index of the alternative/node to make
    continuous.
    * otherwise we assume you want original weights to be the p0 value, and return the parameter `orig`
    :param limit_matrix_calc: A function which takes a single input, the matrix to take the limit of.
    :return: A pandas.Series whose indices are influence_nodes and whose values are the marginal influence
    scores of those nodes with respect to the given row.
    '''
    n = len(mat)
    if influence_nodes is None:
        influence_nodes = [i for i in range(n) if i != row]
    orig_lim = calculus(mat)
    orig_pri = priority_from_limit(orig_lim)[influence_nodes]
    orig_pri /= sum(abs(orig_pri))
    if not p0mode_is_direct(p0mode):
        raise ValueError("p0mode must be a direct p0 value for marginal influence")
    else:
        p0 = p0mode
    if left_or_right <= 0:
        #Calculate left deriv
        new_mat = row_adjust(mat, row, p0-delta, cluster_nodes, p0mode=p0mode)
        lim = limit_matrix_calc(new_mat)
        pri = priority_from_limit(lim)[influence_nodes]
        pri /= sum(abs(pri))
        left_deriv = (pri - orig_pri) / -delta
        if left_or_right < 0:
            #Only want left deriv
            rval = pd.Series(data = left_deriv, index=influence_nodes)
            return rval
    if left_or_right >= 0:
        # Calculate left deriv
        new_mat = row_adjust(mat, row, p0 + delta, cluster_nodes, p0mode=p0)
        lim = limit_matrix_calc(new_mat)
        pri = priority_from_limit(lim)[influence_nodes]
        pri /= sum(abs(pri))
        right_deriv = (pri - orig_pri) / delta
        if left_or_right > 0:
            # Only want right deriv
            rval = pd.Series(data = right_deriv, index=influence_nodes)
            return rval
    #If we make it here, we want avg
    rval = pd.Series(data=(left_deriv + right_deriv)/2, index=influence_nodes)
    return rval

def influence_table(mat, row, pvals=None, cluster_nodes=None, influence_nodes=None, p0mode=None, limit_matrix_calc=calculus, graph=True):
    '''
    Calculates the direct influence score, i.e. it calculates anp row sensitivity for each of pvals values and
    stores the new scores of the influence_nodes.
    :param mat: The scaled supermatrix to perform the calculation on
    :param row: The row to use for anp row sensitivity
    :param pvals: The values to set p to, this should be a list (or list like) object of values before 0 and 1.
    :param cluster_nodes: If you wish to normalize by cluster, this should be the indices of the nodes that are
    in row's cluster (including row itself).
    :param influence_nodes: The indices of the nodes to calculate the influence of, with respect to row. If None
    it calculates the influence of all nodes other than row.
    :param p0mode: Determines how p0 is calculated and has 3 cases:
    * if it is a float, you are directly setting the p0 value to whatever p0mode is
    * if it is an integer, this is the smart p0 mode, and it treats p0mode as the index of the alternative/node to make
    continuous.
    * otherwise we assume you want original weights to be the p0 value, and return the parameter `orig`
    :param limit_matrix_calc:
    :param limit_matrix_calc: A function which takes a single input, the matrix to take the limit of.
    :param graph: If True, we return a matplotlib graph, otherwise we return pandas.DataFrame, p0vals
    :return: If graph=True, we return nothing, but create a matplotlib object and call plt.show().  Otherwise
    we return a pair of items.  The first is the dataframe of results, whose indices are "Node 1", "Node 2", ...
    which corresponds to influence_nodes (and the indices after "Node " are the influence_node indices)
    and has 2 columns, 'x' is the pvals and 'y' is the resulting influence
    score (i.e. changed priority).  The second element is a pd.Series whose indices is the same as the dataframe
    and whose values are pairs of items (x,y) where x is the p0 value for the given alternative and the y is the
    influence score of that alternative at that p-value.
    '''
    if pvals is None:
        xs = [i / 50 for i in range(1, 50)]
    else:
        xs = pvals
    n = len(mat)
    if influence_nodes is None:
        influence_nodes = [i for i in range(n) if i != row]
    df = pd.DataFrame()
    pvals = pd.Series()
    df['x']=xs
    for alt in influence_nodes:
        ys = []
        if isinstance(p0mode, int):
            # This means p0mode is smart, and we should do it smart wrt the alt
            p0mode = alt
        for x in xs:
            new_mat = row_adjust(mat, row, x, cluster_nodes=cluster_nodes, p0mode=p0mode)
            new_lmt = limit_matrix_calc(new_mat)
            new_pri = priority_from_limit(new_lmt)
            new_pri[row] = 0
            new_pri /= sum(new_pri)
            y = new_pri[alt]
            ys.append(y)
        label = "Node " + str(alt)
        p0 = calcp0(mat, row, cluster_nodes, mat[row, alt], p0mode)
        x = p0
        y = linear_interpolate(xs, ys, x)
        if graph:
            plt.plot(xs, ys, label=label)
            plt.scatter(x, y, label=label+" p0")
        else:
            df[label]=ys
            pvals[label]=(x, y)
    if graph:
        plt.legend()
        plt.show()
    else:
        return df, pvals

def influence_table_plot(df, p0s):
    '''
    Graphs the return value of influence_table(graph=False), useful if you want to have both the graph done
    and also the table of values.
    :param df: The 1st returned component from influence_table(graph=False): a dataframe
    :param p0s: The 2nd returned component from influence_table(graph=False): a Series of (x,y)'s
    :return: Nothing, but does call plt.show() to make the matplotlib graph visible.
    '''
    xs = df[df.columns[0]]
    for col, p0 in zip(df.columns[1:], p0s):
        plt.plot(xs, df[col], label=str(col))
        plt.scatter(p0[0], p0[1], label=str(col)+" p0")
    plt.legend()
    plt.show()

def influence_limit(mat, row, cluster_nodes=None, influence_nodes=None, delta=1e-6, p0mode=0.5, limit_matrix_calc=calculus):
    '''
    Calculates the limit influence score of the influence_nodes with respect to row.
    :param mat: The scaled supermatrix to perform the calculation on
    :param row: The row to use for anp row sensitivity
    :param cluster_nodes: If you wish to normalize by cluster, this should be the indices of the nodes that are
    in row's cluster (including row itself).
    :param influence_nodes: The indices of the nodes to calculate the influence of, with respect to row. If None
    it calculates the influence of all nodes other than row.
    :param delta: We use 1-delta for the p-value to plugin to approximate the limit as p -> 1
    :param p0mode: Determines how p0 is calculated and has 3 cases:
    * if it is a float, you are directly setting the p0 value to whatever p0mode is
    * if it is an integer, this is the smart p0 mode, and it treats p0mode as the index of the alternative/node to make
    continuous.
    * otherwise we assume you want original weights to be the p0 value, and return the parameter `orig`
    :param limit_matrix_calc: A function which takes a single input, the matrix to take the limit of.
    :return: A tuple of 2 items, the first is a pandas.Series whose indices are 'Node 1', 'Node 2'
     (and the indices after "Node " are the influence_node indices)
    and whose values are the limit value.  The second element of the returned tuple is a pandas.Series
    with the same indices and whose values are the p0 values we used for that alternative.
    '''
    if not p0mode_is_direct(p0mode):
        raise ValueError("p0mode must be a direct p0 value for limit influence")
    n = len(mat)
    if influence_nodes is None:
        influence_nodes = [i for i in range(n) if i != row]
    df = pd.DataFrame()
    limits = pd.Series()
    p0 = 1 - delta
    p0s = pd.Series()
    for alt in influence_nodes:
        if isinstance(p0mode, int):
            # This means p0mode is smart, and we should do it smart wrt the alt
            p0mode = alt
        new_mat = row_adjust(mat, row, p0, cluster_nodes=cluster_nodes, p0mode=p0mode)
        new_lmt = limit_matrix_calc(new_mat)
        new_pri = priority_from_limit(new_lmt)
        row_pri = new_pri[row]
        new_pri[row] = 0
        new_pri /= sum(new_pri)
        new_pri[row]=row_pri
        y = new_pri[alt]
        label = "Node " + str(alt)
        limits[label]=y
        p0 = calcp0(mat, row, cluster_nodes, mat[row, alt], p0mode)
        p0s[label]=p0
    return limits, p0s

def row_adjust_priority(mat, row, p, cluster_nodes=None, p0mode=None, limit_matrix_calc=calculus,
                        normalize_to_orig=False):
    '''
    Adjusts a row of matrix and recalculates the priorities of all the nodes.
    :param mat: The scaled supermatrix to perform the calculation on
    :param row: The row to use for anp row sensitivity
    :param cluster_nodes: If you wish to normalize by cluster, this should be the indices of the nodes that are
    in row's cluster (including row itself).
    :param p0mode: Determines how p0 is calculated and has 3 cases:
    * if it is a float, you are directly setting the p0 value to whatever p0mode is
    * if it is an integer, this is the smart p0 mode, and it treats p0mode as the index of the alternative/node to make
    continuous.
    * otherwise we assume you want original weights to be the p0 value, and return the parameter `orig`
    :param limit_matrix_calc: A function which takes a single input, the matrix to take the limit of.
    :param normalize_to_orig: If True we normalize the returning priority score so that the [row] index of it has
    the same value as the original and the other values are rescaled.  Otherwise we simply normalize the priority
    vector directly.
    '''
    if normalize_to_orig:
        old_lmt = limit_matrix_calc(mat)
        old_pri = priority_from_limit(old_lmt)
        old_val = old_pri[row]
        old_sum = sum(old_pri) - old_val
    else:
        old_sum = 1
    new_mat = row_adjust(mat, row, p, cluster_nodes=cluster_nodes, p0mode=p0mode)
    new_lmt = limit_matrix_calc(new_mat)
    new_pri = priority_from_limit(new_lmt)
    row_pri = new_pri[row]
    new_pri[row] = 0
    new_pri *= old_sum / sum(new_pri)
    new_pri[row]=row_pri
    return new_pri

def influence_fixed(mat, row, cluster_nodes=None, influence_nodes=None, delta=0.25, p0mode=0.5,
                    limit_matrix_calc=calculus):
    '''
    Calculates fixed influence, i.e. we do row sensitivity and calculate the difference
    :param mat: The scaled supermatrix to perform the calculation on
    :param row: The row to use for anp row sensitivity
    :param cluster_nodes: If you wish to normalize by cluster, this should be the indices of the nodes that are
    in row's cluster (including row itself).
    :param influence_nodes: The indices of the nodes to calculate the influence of, with respect to row. If None
    it calculates the influence of all nodes other than row.
    :param delta: How much to change from p0 for the fixed influence
    :param p0mode: Determines how p0 is calculated and has 3 cases:  NOTE: ONLY DIRECT MODE ALLOWED
    * if it is a float, you are directly setting the p0 value to whatever p0mode is
    * if it is an integer, this is the smart p0 mode, and it treats p0mode as the index of the alternative/node to make
    continuous.
    * otherwise we assume you want original weights to be the p0 value, and return the parameter `orig`
    :param limit_matrix_calc:
    :return: A pandas.Series whose index is influence_nodes and whose values are the influence scores of those nodes
    with respect to the row.
    '''
    if not p0mode_is_direct(p0mode):
        raise ValueError("p0mode must be a direct p0 value for fixed distance influence")
    if influence_nodes is None:
        n = len(mat)
        influence_nodes = [i for i in range(n) if i != row]
    p0 = p0mode + delta
    old_pri = row_adjust_priority(mat, row, 0.5, cluster_nodes, p0mode=0.5, limit_matrix_calc=limit_matrix_calc)
    new_pri = row_adjust_priority(mat, row,p0, cluster_nodes, p0mode=p0mode, limit_matrix_calc=limit_matrix_calc)
    diff = new_pri - old_pri
    rval = pd.Series(data=diff[influence_nodes], index=influence_nodes)
    return rval

def rank_change(vec1, vec2, places_to_rank, rank_change_places=None, round_to_decimal=5):
    '''
    A calculation that rounds 2 vectors to round_to_decimal places and then looks to see if
    the ranking of rounded vec1 is different from rounded vec2.
    :param vec1: A list or list-like object to check rank changing
    :param vec2: Another list or list-like object to check rank changing
    :param places_to_rank: Indices to rank.
    :param rank_change_places: Of the indices we are ranking, which ones are we checking for a change (if None we
    check all indices for rank change).
    :param round_to_decimal: The number of decimal places to round to, before checking for rank changes
    :return: True if a rank change happen and False otherwise
    '''
    if (rank_change_places is None):
        rank_change_places = places_to_rank
    elif isinstance(rank_change_places, int):
        rank_change_places = [rank_change_places]
    vec1 = np.round(vec1, round_to_decimal)
    vec2 = np.round(vec2, round_to_decimal)
    if len(rank_change_places) == 1:
        rk1 = rankdata(vec1)[rank_change_places[0]]
        rk2 = rankdata(vec2)[rank_change_places[0]]
        return rk1 != rk2
    else:
        vec1 = vec1[rank_change_places]
        vec2 = vec2[rank_change_places]
        rk1 = rankdata(vec1)
        rk2 = rankdata(vec2)
        return any(rk1 != rk2)

def influence_rank(mat, row, cluster_nodes=None, influence_nodes=None,
                   limit_matrix_calc=calculus, rank_change_nodes=None, error=1e-5, upper_lower_both=0,
                   round_to_decimal=5, return_full_info=False):
    '''
    Calculates rank influence scores.

    :param mat: The scaled supermatrix to perform the calculation on
    :param row: The row to use for anp row sensitivity
    :param cluster_nodes: If you wish to normalize by cluster, this should be the indices of the nodes that are
    in row's cluster (including row itself).
    :param influence_nodes: The indices of the nodes to calculate the influence of, with respect to row. If None
    it calculates the influence of all nodes other than row.
    :param limit_matrix_calc: A function with one parameter, that calculates the limit matrix.
    :param rank_change_nodes:  The nodes to look for rank change at
    :param error: While we narrow down our search for the p-value that causes a rank change, how close do we want
    the values between a change happening and not, to be.
    :param upper_lower_both:  Do we want to:
    * look for rank change only by changing p > 0.5=p0, if so upper_lower_both > 0
    * look for rank change only by changing p < 0.5=p0, if so upper_lower_both < 0
    * look for rank change by changing p>0.5 and p<0.5, if so upper_lower_both = 0
    :param round_to_decimal:  How many decimals should we round the score to for ranking purposes
    :param return_full_info: If True returns more info, see the return section for more details
    :return: A list of one or more numbers, controlled by return_full_info and upper_lower_both:
    * If upper_lower_both < 0: our p-val search will only be for pval < 0.5
      * If return_full_info is True:
        We return pval_where_rank_chg_happens, score_of_that_pval
      * Otherwise:
        We return score_of_that_pval
    * If upper_lower_both > 0: our pval search will only be for pval > 0.5
      * If return_full_info is True:
        We return pval_where_rank_chg_happens, score_of_that_pval
      * Otherwise:
        We return score_of_that_pval
    * If upper_lower_both = 0: we check both lower and upper
      * If return_full_info is True:
        We return max_of_upper_lower_scores, lower_rank_value, lower_rank_chg_score, upper_rank_value, upper_rank_chg_score
      * Otherwise we return:
        max_of_upper_lower_scores
    '''
    n = len(mat)
    if influence_nodes is None:
        influence_nodes = [i for i in range(n) if i != row]
    if rank_change_nodes is None:
        rank_change_nodes = influence_nodes
    #Start with upper rank change
    if upper_lower_both >= 0:
        #Initial bounds
        lower = 0.5
        upper = 0.99999
        p0mode = 0.5
        lower_pri = row_adjust_priority(mat, row, lower, cluster_nodes, p0mode=p0mode, limit_matrix_calc=limit_matrix_calc)
        upper_pri = row_adjust_priority(mat, row, upper, cluster_nodes, p0mode=p0mode, limit_matrix_calc=limit_matrix_calc)
        if not rank_change(lower_pri, upper_pri, influence_nodes, rank_change_nodes, round_to_decimal):
            # There is no rank change to start with
            upper_rank_chg = 1
        else:
            while (upper - lower) > error:
                mid = (upper + lower)/2
                mid_pri = row_adjust_priority(mat, row, mid, cluster_nodes, p0mode=p0mode,
                                              limit_matrix_calc=limit_matrix_calc)
                if rank_change(lower_pri, mid_pri, influence_nodes, rank_change_nodes, round_to_decimal):
                    upper = mid
                    upper_pri = mid_pri
                elif rank_change(mid_pri, upper_pri, influence_nodes, rank_change_nodes, round_to_decimal):
                    lower = mid
                    lower_pri = mid_pri
                else:
                    # This should not happen
                    raise ValueError("Please report this error: rank influence impossible state")
            upper_rank_chg = mid
        upper_rank_value = (1 - upper_rank_chg) / (1 - 0.5)
    if upper_lower_both <= 0:
        #Initial bounds
        lower = 0.00001
        upper = 0.5
        p0mode = 0.5
        lower_pri = row_adjust_priority(mat, row, lower, cluster_nodes, p0mode=p0mode, limit_matrix_calc=limit_matrix_calc)
        upper_pri = row_adjust_priority(mat, row, upper, cluster_nodes, p0mode=p0mode, limit_matrix_calc=limit_matrix_calc)
        if not rank_change(lower_pri, upper_pri, influence_nodes, rank_change_nodes, round_to_decimal):
            # There is no rank change to start with
            lower_rank_chg = 0.0
        else:
            while (upper - lower) > error:
                mid = (upper + lower)/2
                mid_pri = row_adjust_priority(mat, row, mid, cluster_nodes, p0mode=p0mode,
                                              limit_matrix_calc=limit_matrix_calc)
                if rank_change(lower_pri, mid_pri, influence_nodes, rank_change_nodes, round_to_decimal):
                    upper = mid
                    upper_pri = mid_pri
                elif rank_change(mid_pri, upper_pri, influence_nodes, rank_change_nodes, round_to_decimal):
                    lower = mid
                    lower_pri = mid_pri
                else:
                    # This should not happen
                    raise ValueError("Please report this error: rank influence impossible state")
            lower_rank_chg = mid
        lower_rank_value = (lower_rank_chg)/(0.5 - 0)
    if upper_lower_both < 0:
        if return_full_info:
            return lower_rank_value, lower_rank_chg
        else:
            return lower_rank_value
    elif upper_lower_both > 0:
        if return_full_info:
            return upper_rank_value, upper_rank_chg
        else:
            return upper_rank_value
    else:
        rval = max(lower_rank_value, upper_rank_value)
        if return_full_info:
            return rval, lower_rank_value, lower_rank_chg, upper_rank_value, upper_rank_chg
        else:
            return rval


'''
Contains all limit matrix calculations
@author Dr. Bill Adams
'''

import numpy as np
from copy import deepcopy


def _mat_pow2(mat, power):
    '''
    Calculates mat ** N where N >= power and N is a power of 2.  It does this by squaring mat, and
    squaring that, etc, until it reaches the desired level.  It takes at most log_2(power)+1 matrix
    multiplications to do this, which is much preferred for large powers.
    :param mat: The numpy array to raise to a power.
    :param power: The power to be greater than or equal to
    :return: The resulting power of the matrix
    '''
    last = deepcopy(mat)
    nextm = deepcopy(mat)
    count=1
    while count <= power:
        np.matmul(last, last, nextm)
        tmp = last
        last = nextm
        nextm = tmp
        count *= 2
    return last

def normalize(mat, inplace=False):
    '''
    Makes the columns of a matrix add to 1 (unless the column summed to zero, in which case it is left unchanged)
    Does this by dividing each column by the sum of that column.
    :param mat: The matrix to normalize
    :param inplace: If true normalizes the matrix sent in, otherwise it leaves that matrix alone, and returns a
    normalized copy
    :return: If inplace=False, it returns the normalized matrix, leaving the param mat unchanged.  Otherwise it
    returns nothing and normalizes the param mat.
    '''
    div = mat.sum(axis=0)
    for i in range(len(div)):
        if div[i] == 0:
            div[i] = 1.0
    if not inplace:
        return mat/div
    else:
        thetype = mat.dtype
        #Let's check that the matrix can do float arith
        old = mat[0,0]
        mat[0,0] = 1./3.
        if mat[0,0] == 0:
            # It is an integer type matrix, which causes a fail
            raise ValueError("Matrix cannot be integer type for inplace normalization.")
        #Reset
        mat[0,0]=old
        np.divide(mat, div, out=mat)

def hiearhcy_formula(mat):
    '''
    Uses the hierarchy formula to calculate the limit matrix.  This is essentially the normalization of
    the sum of higher powers of mat.
    :param mat: A square nump.array that you wish to find the limit matrix of, using the hiearchy formula.
    :return: The limit matrix, unless the matrix was not a hiearchy.  If the matrix was not a hierarchy we return None
    '''
    size = len(mat)
    big = _mat_pow2(mat, size+1)
    if np.count_nonzero(big) != 0:
        # Not a hiearchy, return None to indicate this
        return None
    summ = deepcopy(mat)
    thispow = deepcopy(mat)
    nextpow = deepcopy(mat)
    for i in range(size-2):
        np.matmul(thispow, mat, nextpow)
        np.add(summ, nextpow, summ)
        tmp = thispow
        thispow = nextpow
        nextpow = tmp
    rval = normalize(summ)
    return rval

def calculus(mat, error=1e-10, max_iters=1000, use_hierarchy_formula=True, col_scale_type=None):
    '''
    Calculates the 'Calculus Type' limit matrix from superdecisions
    :param mat:
    :param error:
    :param max_iters:
    :param use_hierarchy_formula:
    :param col_scale_type: A string if 'all' it scales mat1 by max(mat1) and similarly for mat2
    otherwise, it scales by column
    :return:
    '''
    size = len(mat)
    diff = 0.0
    start_pow = size+10
    start = _mat_pow2(mat, start_pow)
    tmp1 = deepcopy(mat)
    tmp2 = deepcopy(mat)
    tmp3 = deepcopy(mat)
    # Now we need matrices to store the intermediate results
    pows = [start]
    for i in range(size - 1):
        # Add next one
        pows.append(np.matmul(mat, pows[-1]))
        diff = normalize_cols_dist(pows[-1], pows[-2], tmp1, tmp2, tmp3, col_scale_type)
        # print(diff)
        if diff < error:
            # Already converged, done
            return pows[-1]
    # print(pows)
    for count in range(max_iters):
        nextp = pows[0]
        np.matmul(pows[-1], mat, nextp)
        # print(pows)
        for i in range(len(pows) - 1):
            pows[i] = pows[i + 1]
        pows[-1] = nextp
        # print(pows)
        # Check convergence
        for i in range(len(pows) - 1):
            diff = normalize_cols_dist(pows[i], nextp, tmp1, tmp2, tmp3, col_scale_type)
            # print(pows[i])
            # print(nextp)
            # print(diff)
            if diff < error:
                return nextp / nextp.sum(axis=0)


def normalize_cols_dist(mat1, mat2, tmp1=None, tmp2=None, tmp3=None, col_scale_type=None):
    '''
    Calculates the distance between matrices mat1 and mat2 after they have
    been column normalized.  tmp1, tmp2, tmp3
    are temporary matrices to store the normalized versions of things.  This
    code could be called many times in a limit matrix calculation, and allocating
    and freeing those temporary storage bits could take a lot of cycles.  This
    way you allocate them once at the top level loop of the limit matrix calculation
    and they are reused again and again.

    If you do not wish to avail yourself of this savings, simply leave them as None's
    and the algorithm will allocate as appropriate
    :param mat1:
    :param mat2:
    :param tmp1: A temporary storage matrix of same size as mat1 and mat2.  If None, it will be allocated inside the fx.
    :param tmp2: A temporary storage matrix of same size as mat1 and mat2.  If None, it will be allocated inside the fx.
    :param tmp3: A temporary storage matrix of same size as mat1 and mat2.  If None, it will be allocated inside the fx.
    :param col_scale_type: A string if 'all' it scales mat1 by max(mat1) and similarly for mat2
    otherwise, it scales by column
    :return: The maximum difference between the column normalized versions of mat1 and mat2
    '''
    tmp1 = tmp1 if tmp1 is not None else deepcopy(mat1)
    tmp2 = tmp2 if tmp2 is not None else deepcopy(mat1)
    tmp3 = tmp3 if tmp3 is not None else deepcopy(mat1)
    if col_scale_type == "all":
        div1 = max(mat1)
        div2 = max(mat2)
    else:
        div1 = mat1.max(axis=0)
        div2 = mat2.max(axis=0)
        for i in range(len(div1)):
            if div1[i]==0:
                div1[i]=1
            if div2[i] == 0:
                div2[i]=1
    np.divide(mat1, div1, tmp1)
    #I think this is wrong, I'm just doing it the way I think it should
    #np.divide(mat2, div2.max(axis=0), tmp2)
    np.divide(mat2, div2, tmp2)
    np.subtract(tmp1, tmp2, tmp3)
    np.absolute(tmp3, tmp3)
    return np.max(tmp3)

def zero_cols(full_mat, non_zero=False):
    '''
    Returns the list of indices of columns that are zero or non_zero
    depending on the parameter non_zero
    :param mat: The matrix to search over
    :param non_zero: If False, returns the indices of columns that are zero, otherwise
    returns indices of columns that a not zero.
    :return: A list of indices of columns of the type determined by the parameter non_zero.
    '''
    size = len(full_mat)
    rval = []
    rval_other = []
    for col in range(size):
        is_zero = True
        for row in range(size):
            if full_mat[row, col] != 0:
                is_zero = False
                break
        if is_zero:
            rval.append(col)
        else:
            rval_other.append(col)
    if non_zero:
        return rval_other
    else:
        return rval

def hierarchy_nodes(mat):
    '''
    Returns the indices of the nodes that are hierarchy ones.  The others are not hierachy
    :param mat: A supermatrix (scaled or non-scaled, both work).
    :return: List of indices that are the nodes which are hierarhical.
    '''
    size = len(mat)
    start_pow = size
    full_mat = _mat_pow2(mat, start_pow)
    return zero_cols(full_mat)

def two_two_breakdown(mat, upper_right_indices):
    '''
    Given the indices for the upper right portion of a matrix, break the matrix
    down into a 2x2 matrix with the submatrices in each piece.  Useful for
    limit matrix calculations that split the problem up into the hierarchical and
    network components and do separate calculations and then bring them together.
    :param mat: The matrix to split into
    A | B
    C | D
    form, where A is the "upper_right_indcies" and D is the opposite
    :param upper_righ_indices: List of indices of the upper right positions.
    :return: A list of [A, B, C, D] of those matrices
    '''
    total_n = len(mat)
    lower_left_indices = [i for i in range(total_n) if i not in upper_right_indices]
    upper_n = len(upper_right_indices)
    lower_n = total_n - upper_n
    A = np.zeros([upper_n, upper_n])
    B = np.zeros([upper_n, lower_n])
    C = np.zeros([lower_n, upper_n])
    D = np.zeros([lower_n, lower_n])
    for i in range(upper_n):
        row = upper_right_indices[i]
        for j in range(upper_n):
            col = upper_right_indices[j]
            A[i,j]=mat[row,col]
        for j in range(lower_n):
            col = lower_left_indices[j]
            B[i,j]=mat[row, col]
    for i in range(lower_n):
        row = lower_left_indices[i]
        for j in range(upper_n):
            col = upper_right_indices[j]
            C[i,j]=mat[row,col]
        for j in range(lower_n):
            col = lower_left_indices[j]
            D[i,j]=mat[row, col]
    return (A, B, C, D)

def limit_sinks(mat, straight_normalizer=True):
    '''
    Performs the limit with sinks calculation.  We break up the matrix
    into sinks and nonsinks, and use those pieces.
    :param mat:
    :param straight_normalizer: If False we normalize at each step, if True
    we normalize at the end.
    :return:
    '''
    n = len(mat)
    nonsinks = zero_cols(mat, non_zero=True)
    sinks = zero_cols(mat, non_zero=False)
    if len(nonsinks) == n:
        # There are no sinks, return calculus type instead
        return calculus(mat)
    # Okay we made it here, we need to get the A and B portions
    (B, z1, A, z2) = two_two_breakdown(mat, nonsinks)
    # Make sure z1 and z2 are zero
    limitB = calculus(B)
    if not straight_normalizer:
        limitB = normalize(limitB)
    axblimit = np.matmul(A, limitB)
    rval = np.zeros([n, n])
    for i in range(len(nonsinks)):
        orig_row = nonsinks[i]
        for j in range(len(nonsinks)):
            orig_col = nonsinks[j]
            rval[orig_row, orig_col] = limitB[i, j]
    for i in range(len(sinks)):
        orig_row = sinks[i]
        for j in range(len(nonsinks)):
            orig_col=nonsinks[j]
            rval[orig_row, orig_col] = axblimit[i, j]
    if straight_normalizer:
        rval = normalize(rval)
    return rval


def limit_newhierarchy(mat, with_limit=False, error=1e-10, col_scale_type = None, max_count = 1000):
    '''
    Performs the new hiearchy limit matrix calculation
    :param mat:
    :param with_limit:
    :return:
    '''
    n = len(mat)
    hier_nodes = hierarchy_nodes(mat)
    net_nodes = [i for i in range(n) if i not in hier_nodes]
    (B, z1, A, C) = two_two_breakdown(mat, net_nodes)
    if len(net_nodes) == n:
        return calculus(mat)
    elif len(hier_nodes) == n:
        return hiearhcy_formula(mat)
    limitB = calculus(B)
    limitC = calculus(C)
    lowerLeftCorner = np.matmul(A, limitB) + np.matmul(B, C)
    lowerLeftCorner = normalize(lowerLeftCorner)
    if with_limit:
        laststep = lowerLeftCorner
        diff = 1
        tmp1 = deepcopy(mat)
        tmp2 = deepcopy(mat)
        tmp3 = deepcopy(mat)
        count = 0
        while (diff > error) and (count < max_count):
            nextstep = np.matmul(A, laststep) + np.matmul(B, C)
            diff = normalize_cols_dist(laststep, nextstep, tmp1, tmp2, tmp3, col_scale_type=col_scale_type)
            laststep = nextstep
            count+=1
        lowerLeftCorner = nextstep
    rval = np.zeros([n, n])
    for i in range(len(net_nodes)):
        orig_row = net_nodes[i]
        for j in range(len(net_nodes)):
            orig_col = net_nodes[j]
            rval[orig_row, orig_col] = limitB[i, j]
    for i in range(len(hier_nodes)):
        orig_row = hier_nodes[i]
        for j in range(len(net_nodes)):
            orig_col=net_nodes[j]
            rval[orig_row, orig_col] = lowerLeftCorner[i, j]
        for j in range(len(hier_nodes)):
            orig_col=hier_nodes[j]
            rval[orig_row, orig_col] = C[i, j]
    rval = normalize(rval)
    return rval

def priority_from_limit(limit_matrix):
    '''
    Calculates the priority from a limit matrix, i.e. sums columns and divides by the number of
    columns.
    :param limit_matrix:
    :return:
    '''
    rval = limit_matrix.sum(axis=1)
    n = len(rval)
    rval /= n
    return rval